/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 2016 OpenFOAM Foundation
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "bslearsm.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace RASModels
{

// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //
const dimensionedScalar smallOmega_ ("smallOmega",dimensionSet (0,0,-1,0,0,0,0), 1e-9);
const dimensionedScalar smallDimless_ ("smallDimless",dimless, 1e-9);

template<class BasicTurbulenceModel>
bslearsm<BasicTurbulenceModel>::bslearsm
(
    const alphaField& alpha,
    const rhoField& rho,
    const volVectorField& U,
    const surfaceScalarField& alphaRhoPhi,
    const surfaceScalarField& phi,
    const transportModel& transport,
    const word& propertiesName,
    const word& type
)
:
    Foam::kOmegaSST
    <
        eddyViscosity<RASModel<BasicTurbulenceModel>>,
        BasicTurbulenceModel
    >
    (
        type,
        alpha,
        rho,
        U,
        alphaRhoPhi,
        phi,
        transport,
        propertiesName
    ),
    alphaK1_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "alphaK1",
            this->coeffDict_,
            0.5
        )
    ),
    alphaOmega2_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "alphaOmega2",
            this->coeffDict_,
            0.85616
        )
    ),
    gamma1_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "gamma1",
            this->coeffDict_,
            0.5532
        )
    ),
    gamma2_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "gamma2",
            this->coeffDict_,
            0.4403
        )
    ),
    Prt_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "Prt",
            this->coeffDict_,
            1.0
        )
    ),
    kappa_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "kappa",
            this->coeffDict_,
            0.41
        )
    ),
    C1_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "C1",
            this->coeffDict_,
            1.8
        )
    ),

    A1_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "A1",
            this->coeffDict_,
            1.245
        )
    ),
    tau_ ( max((1./(this->betaStar_*this->omega_)),6.*sqrt( this->nu() / (this->betaStar_*this->k_*this->omega_)) ) ),
    gradU_( fvc::grad(this->U_)-(1./3.)*I*tr(fvc::grad(this->U_)) ),
    S_( tau_*symm(gradU_.T())),
    W_( tau_*skew(gradU_.T())),

    IIs_( tr(S_&S_)),
    IIo_( tr(W_&W_)),
    IV_( tr((S_&W_)&W_)),
    C1prime_( (9./4.)*(C1_-1.)),           //constant
    P1_( C1prime_*((sqr(C1prime_)/27.)+(9.*IIs_/20.)-(2.*IIo_/3.))),
    P2_( sqr(P1_)-pow(((sqr(C1prime_)/9.)+(9.*IIs_/10.)+(2.*IIo_/3.)),3)),
    N_(
        IOobject
        (
            "N",
            this->runTime_.timeName(),
            this->mesh_,
            IOobject::NO_READ,
            IOobject::AUTO_WRITE
        ),
        C1prime_+(9./4.)*sqrt(2.*this->betaStar_*IIs_)
    ),

//    N_(scalar(0.)*IIs_),
    Q_( (sqr(N_)-2.*IIo_)/A1_),
    Q1_( Q_*(2.*sqr(N_)-IIo_)/6.),
    Beta1_( -1.*N_/(Q_+smallDimless_)),
    Beta2_( 0.*N_),         //0 work around
    Beta3_( -2.*IV_/(N_*Q1_+smallDimless_)),
    Beta4_( -1./(Q_+smallDimless_)),
    Beta6_( -1.*N_/(Q1_+smallDimless_)),
    Beta9_( 1./(Q1_+smallDimless_)),
    check0_(tr(S_)),        //T1 term
    check1_(tr(((S_ & S_)-(1./3.)*IIs_*I))),        //T2 term
    check2_(tr(((W_ & W_)-(1./3.)*IIo_*I))),        //T3 term
    check3_(tr(((S_ & W_)-(W_ & S_)))),             //T4 term
    check4_(tr((((S_ & W_) & W_) + ((W_ & W_) & S_) - (2./3.)*IV_*I - IIo_*S_))),   //T6 term
    check5_(tr((((W_ & S_) & (W_ & W_)) - ((W_ & W_) & (S_ & W_)) + 0.5*IIo_*((S_ & W_) - (W_ & S_))))),  //T9 term
    aij_(
        IOobject
        (
            "aij",
            this->runTime_.timeName(),
            this->mesh_,
            IOobject::NO_READ,
            IOobject::AUTO_WRITE
        ),
            symm(
            Beta1_*S_
            + Beta3_*((W_ & W_)-(1./3.)*IIo_*I)
            + Beta4_*((S_ & W_)-(W_ & S_))
            + Beta6_*(((S_ & W_) & W_) + ((W_ & W_) & S_) - (2./3.)*IV_*I - IIo_*S_)
            )
        ),
    tauij_(
        "tauij",
        symm(
            this->k_*(aij_+(2./3.)*I)
            )
        ),
    Nij_(
        "Nij",
        symm(

            (this->k_*aij_+2.0*(this->nut_)*symm(gradU_))*this->rho_

            )
        )
{
    if (type == typeName)
    {
        this->printCoeffs(type);
        Info << "BSL-EARSM Turbulence Model." << endl;
    }
}

template<class BasicTurbulenceModel>
tmp<fvVectorMatrix> bslearsm<BasicTurbulenceModel>::divDevRhoReff(volVectorField& U) const
{
    //Info << "Using divDevRhoReff from BSL-EARSM" << endl;
    return
    (
        fvc::div(this->alpha_*Nij_)
 	  - fvc::div((this->alpha_*this->rho_*this->nuEff())*dev2(T(fvc::grad(U))))
 	  - fvm::laplacian(this->alpha_*this->rho_*this->nuEff(), U)
    );
}

template<class BasicTurbulenceModel>
void bslearsm<BasicTurbulenceModel>::correct()
{
    if (!this->turbulence_)
    {
        this->nut_ = this->k_/(this->omega_+smallOmega_);
        this->nut_.correctBoundaryConditions();
        BasicTurbulenceModel::correctNut();
        return;
    }

    // Local references
    const alphaField& alpha = this->alpha_;
    const rhoField& rho = this->rho_;
    const surfaceScalarField& alphaRhoPhi = this->alphaRhoPhi_;
    const volVectorField& U = this->U_;
    volScalarField& nut = this->nut_;
    fv::options& fvOptions(fv::options::New(this->mesh_));

    BasicTurbulenceModel::correct();

    volScalarField divU(fvc::div(fvc::absolute(this->phi(), U)));

    tmp<volTensorField> tgradU = fvc::grad(U);


    tau_ = max((1./(this->betaStar_*this->omega_)),6.*sqrt(this->nu()/(this->betaStar_*this->k_*this->omega_)));

    gradU_ = fvc::grad(this->U_)-(1./3.)*I*tr(fvc::grad(this->U_));
    S_ = tau_*symm(gradU_.T());
    W_ = tau_*skew(gradU_.T());

    IIs_ = tr(S_ & S_);
    IIo_ = tr(W_ & W_);
    IV_  = tr((S_ & W_) & W_);
    C1prime_ = (9./4.)*(C1_-1.);           //constant
    P1_ = C1prime_*((sqr(C1prime_)/27.)+(9.*IIs_/20.)-(2.*IIo_/3.));
    P2_ = sqr(P1_)-pow(((sqr(C1prime_)/9.)+(9.*IIs_/10.)+(2.*IIo_/3.)),3);
    forAll(N_,cellI)
    {
        if(P2_[cellI]  >= 0)
        {
        N_[cellI]= (C1prime_.value()/3.)+pow((P1_[cellI]+sqrt(P2_[cellI])),(1./3.))
                    +sign(P1_[cellI]-sqrt(P2_[cellI]))*pow(mag(P1_[cellI]-sqrt(P2_[cellI])),(1./3.));
        }
        else
        {
        N_[cellI]= (C1prime_.value()/3.)+2.*pow((sqr(P1_[cellI])-P2_[cellI]),(1./6.))
                    *cos((1./3.)*acos(P1_[cellI]/(sqrt(sqr(P1_[cellI])-P2_[cellI]))));
        }
    }
    //N_.correctBoundaryConditions(); 
    Q_ = (sqr(N_)-2.*IIo_)/A1_;
    Q1_ = Q_*(2.*sqr(N_)-IIo_)/6.;
    Beta1_ = -1.*N_/(Q_+smallDimless_);
    Beta2_ = 0.*N_;     //0 work around
    Beta3_ = -2.*IV_/(N_*Q1_+smallDimless_);
    Beta4_ = -1./(Q_+smallDimless_);
    Beta6_ = -1.*N_/(Q1_+smallDimless_);
    Beta9_ = 1./(Q1_+smallDimless_);

    check0_=tr(S_);                             //T1 term
    check1_=tr(((S_ & S_)-(1./3.)*IIs_*I));        //T2 term
    check2_=tr(((W_ & W_)-(1./3.)*IIo_*I));       //T3 term
    check3_=tr(((S_ & W_)-(W_ & S_)));             //T4 term
    check4_=tr((((S_ & W_) & W_) + ((W_ & W_) & S_) - (2./3.)*IV_*I - IIo_*S_));   //T6 term
    check5_=tr((((W_ & S_) & (W_ & W_)) - ((W_ & W_) & (S_ & W_)) + 0.5*IIo_*((S_ & W_) - (W_ & S_))));  //T9 term
    //Info<< "div(U_)  "<<" max:"<< max(fvc::div(this->U_)).value() <<
    //" min:"<< min(fvc::div(this->U_)).value() <<" avg:"<< average(fvc::div(this->U_)).value() << endl;
    Info<< "div(phi_)  "<<" max:"<< max(fvc::div(this->phi())).value() <<
    " min:"<< min(fvc::div(this->phi())).value() <<" avg:"<< average(fvc::div(this->phi())).value() << endl;
    Info<< "gradU_ TRACE  "<<" max:"<< max(tr(gradU_)).value() <<
    " min:"<< min(tr(gradU_)).value() <<" avg:"<< average(tr(gradU_)).value() << endl;
    Info<< "T1 TRACE  "<<" max:"<< max(check0_).value() <<
    " min:"<< min(check0_).value() <<" avg:"<< average(check0_).value() << endl;
    Info<< "T3 TRACE  "<<" max:"<< max(check2_).value() <<
    " min:"<< min(check2_).value() <<" avg:"<< average(check2_).value() << endl;
    Info<< "T4 TRACE  "<<" max:"<< max(check3_).value() <<
    " min:"<< min(check3_).value() <<" avg:"<< average(check3_).value() << endl;
    Info<< "T6 TRACE  "<<" max:"<< max(check4_).value() <<
    " min:"<< min(check4_).value() <<" avg:"<< average(check4_).value() << endl;
    Info<< "N        "<<" max:"<< max(N_).value() <<
    " min:"<< min(N_).value() <<" avg:"<< average(N_).value() << endl;

    aij_ = symm(
            Beta1_*S_
            //+ Beta2_*((S_ & S_)-(1./3.)*IIs_*I)
            + Beta3_*((W_ & W_)-(1./3.)*IIo_*I)
            + Beta4_*((S_ & W_)-(W_ & S_))
            + Beta6_*(((S_ & W_) & W_) + ((W_ & W_) & S_) - (2./3.)*IV_*I - IIo_*S_)
            //+ Beta9_*(((W_ & S_) & (W_ & W_)) - ((W_ & W_) & (S_ & W_)) + 0.5*IIo_*((S_ & W_) - (W_ & S_)))
            );
    Info<< "aij_ TRACE  "<<" max:"<< max(tr(aij_)).value() <<
    " min:"<< min(tr(aij_)).value() <<" avg:"<< average(tr(aij_)).value() << endl;

    tauij_ = symm(this->k_*(aij_+(2./3.)*I));

    Nij_ = symm(this->k_*aij_+2.0*this->nut_*symm(gradU_))*this->rho_;       //Non-linear part of Reynolds Stresses    


    volScalarField G(this->GName(), -(tauij_&&gradU_));
    tgradU.clear();

    // Update omega and G at the wall
    this->omega_.boundaryFieldRef().updateCoeffs();

    volScalarField CDkOmega
    (
        (2*this->alphaOmega2_)*(fvc::grad(this->k_) & fvc::grad(this->omega_))/this->omega_
    );

    volScalarField F1(this->F1(CDkOmega));
    {
        volScalarField gamma(this->gamma(F1));
        volScalarField beta(this->beta(F1));

        // Turbulent frequency equation
        tmp<fvScalarMatrix> omegaEqn
        (
            fvm::ddt(alpha, rho, this->omega_)
          + fvm::div(alphaRhoPhi, this->omega_)
          - fvm::laplacian(alpha*rho*this->DomegaEff(F1), this->omega_)
         ==
            alpha*rho*gamma
           *min
            (
                G,
                this->c1_*this->betaStar_*this->k_*this->omega_
            )*this->omega_/this->k_
          - fvm::SuSp((2.0/3.0)*alpha*rho*gamma*divU, this->omega_)
          - fvm::Sp(alpha*rho*beta*this->omega_, this->omega_)
          - fvm::SuSp
            (
                alpha*rho*(F1 - scalar(1))*CDkOmega/this->omega_,
                this->omega_
            )
        );

        omegaEqn.ref().relax();
        fvOptions.constrain(omegaEqn.ref());
        omegaEqn.ref().boundaryManipulate(this->omega_.boundaryFieldRef());
        solve(omegaEqn);
        fvOptions.correct(this->omega_);
        bound(this->omega_, this->omegaMin_);
    }

    // Turbulent kinetic energy equation
    tmp<fvScalarMatrix> kEqn
    (
        fvm::ddt(alpha, rho, this->k_)
      + fvm::div(alphaRhoPhi, this->k_)
      - fvm::laplacian(alpha*rho*this->DkEff(F1), this->k_)
     ==
        min(alpha*rho*G, (this->c1_*this->betaStar_)*alpha*rho*this->k_*this->omega_)
      - fvm::SuSp((2.0/3.0)*alpha*rho*divU, this->k_)
      - fvm::Sp(alpha*rho*this->betaStar_*this->omega_,this->k_)
    );

    kEqn.ref().relax();
    fvOptions.constrain(kEqn.ref());
    solve(kEqn);
    fvOptions.correct(this->k_);
    bound(this->k_, this->kMin_);

    //this->correctNut(S2, F23);
    nut = this->k_/(this->omega_+smallOmega_);
    nut.correctBoundaryConditions();
    BasicTurbulenceModel::correctNut();
}
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace RASModels
} // End namespace Foam

// ************************************************************************* //
